# -*- coding: utf-8 -*-
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import unittest
from trytond.tests.test_tryton import activate_module, with_transaction
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.exceptions import UserError

_context = {'_check_access': True}


class ModelAccessTestCase(unittest.TestCase):
    'Test Model Access'

    @classmethod
    def setUpClass(cls):
        activate_module('tests')

    @with_transaction(context=_context)
    def test_perm_read(self):
        'Test Read Access'
        pool = Pool()
        ModelAccess = pool.get('ir.model.access')
        TestAccess = pool.get('test.access')
        Model = pool.get('ir.model')
        Group = pool.get('res.group')

        model, = Model.search([('model', '=', 'test.access')])

        test, = TestAccess.create([{}])

        # Without model access
        TestAccess.read([test.id])

        # With model access

        # One access allowed for any group
        model_access_wo_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': None,
                    'perm_read': True,
                    }])
        TestAccess.read([test.id])

        # One access disallowed for any group
        ModelAccess.write([model_access_wo_group], {
                'perm_read': False,
                })
        self.assertRaises(UserError, TestAccess.read, [test.id])

        # Two access rules with one group allowed
        group, = Group.search([('users', '=', Transaction().user)])
        model_access_w_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': group.id,
                    'perm_read': True,
                    }])

        TestAccess.read([test.id])

        # Two access rules with both allowed
        ModelAccess.write([model_access_wo_group], {
                'perm_read': True,
                })
        TestAccess.read([test.id])

        # Two access rules with any group allowed
        ModelAccess.write([model_access_w_group], {
                'perm_read': False,
                })
        TestAccess.read([test.id])

        # Two access rules with both disallowed
        ModelAccess.write([model_access_wo_group], {
                'perm_read': False,
                })
        self.assertRaises(UserError, TestAccess.read, [test.id])

        # One access disallowed for one group
        ModelAccess.delete([model_access_wo_group])
        self.assertRaises(UserError, TestAccess.read, [test.id])

        # One access allowed for one group
        ModelAccess.write([model_access_w_group], {
                'perm_read': True,
                })
        TestAccess.read([test.id])

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        ModelAccess.write([model_access_w_group], {
                'group': group.id,
                })
        TestAccess.read([test.id])

        # One access disallowed for one other group
        ModelAccess.write([model_access_w_group], {
                'perm_read': False,
                })
        TestAccess.read([test.id])

    @with_transaction(context=_context)
    def test_perm_write(self):
        'Test Write Access'
        pool = Pool()
        ModelAccess = pool.get('ir.model.access')
        TestAccess = pool.get('test.access')
        Model = pool.get('ir.model')
        Group = pool.get('res.group')

        model, = Model.search([('model', '=', 'test.access')])

        test, = TestAccess.create([{}])

        # Without model access
        TestAccess.write([test], {})

        # With model access

        # One access allowed for any group
        model_access_wo_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': None,
                    'perm_write': True,
                    }])
        TestAccess.write([test], {})

        # One access disallowed for any group
        ModelAccess.write([model_access_wo_group], {
                'perm_write': False,
                })
        self.assertRaises(UserError, TestAccess.write, [test], {})

        # Two access rules with one group allowed
        group, = Group.search([('users', '=', Transaction().user)])
        model_access_w_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': group.id,
                    'perm_write': True,
                    }])
        TestAccess.write([test], {})

        # Two access rules with both allowed
        ModelAccess.write([model_access_wo_group], {
                'perm_write': True,
                })
        TestAccess.write([test], {})

        # Two access rules with any group allowed
        ModelAccess.write([model_access_w_group], {
                'perm_write': False,
                })
        TestAccess.write([test], {})

        # Two access rules with both disallowed
        ModelAccess.write([model_access_wo_group], {
                'perm_write': False,
                })
        self.assertRaises(UserError, TestAccess.write, [test], {})

        # One access disallowed for one group
        ModelAccess.delete([model_access_wo_group])
        self.assertRaises(UserError, TestAccess.write, [test], {})

        # One access allowed for one group
        ModelAccess.write([model_access_w_group], {
                'perm_write': True,
                })
        TestAccess.write([test], {})

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        ModelAccess.write([model_access_w_group], {
                'group': group.id,
                })
        TestAccess.write([test], {})

        # One access disallowed for one other group
        ModelAccess.write([model_access_w_group], {
                'perm_write': False,
                })
        TestAccess.write([test], {})

    @with_transaction(context=_context)
    def test_perm_create(self):
        'Test Create Access'
        pool = Pool()
        ModelAccess = pool.get('ir.model.access')
        TestAccess = pool.get('test.access')
        Model = pool.get('ir.model')
        Group = pool.get('res.group')

        model, = Model.search([('model', '=', 'test.access')])

        # Without model access
        TestAccess.create([{}])

        # With model access

        # One access allowed for any group
        model_access_wo_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': None,
                    'perm_create': True,
                    }])
        TestAccess.create([{}])

        # One access disallowed for any group
        ModelAccess.write([model_access_wo_group], {
                'perm_create': False,
                })
        self.assertRaises(UserError, TestAccess.create, {})

        # Two access rules with one group allowed
        group, = Group.search([('users', '=', Transaction().user)])
        model_access_w_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': group.id,
                    'perm_create': True,
                    }])

        TestAccess.create([{}])

        # Two access rules with both allowed
        ModelAccess.write([model_access_wo_group], {
                'perm_create': True,
                })
        TestAccess.create([{}])

        # Two access rules with any group allowed
        ModelAccess.write([model_access_w_group], {
                'perm_create': False,
                })
        TestAccess.create([{}])

        # Two access rules with both disallowed
        ModelAccess.write([model_access_wo_group], {
                'perm_create': False,
                })
        self.assertRaises(UserError, TestAccess.create, [{}])

        # One access disallowed for one group
        ModelAccess.delete([model_access_wo_group])
        self.assertRaises(UserError, TestAccess.create, [{}])

        # One access allowed for one group
        ModelAccess.write([model_access_w_group], {
                'perm_create': True,
                })
        TestAccess.create([{}])

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        ModelAccess.write([model_access_w_group], {
                'group': group.id,
                })
        TestAccess.create([{}])

        # One access disallowed for one other group
        ModelAccess.write([model_access_w_group], {
                'perm_create': False,
                })
        TestAccess.create([{}])

    @with_transaction(context=_context)
    def test_perm_delete(self):
        'Test Delete Access'
        pool = Pool()
        ModelAccess = pool.get('ir.model.access')
        TestAccess = pool.get('test.access')
        Model = pool.get('ir.model')
        Group = pool.get('res.group')

        model, = Model.search([('model', '=', 'test.access')])

        tests = [TestAccess.create([{}])[0] for x in range(11)]

        # Without model access
        TestAccess.delete([tests.pop()])

        # With model access

        # One access allowed for any group
        model_access_wo_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': None,
                    'perm_delete': True,
                    }])
        TestAccess.delete([tests.pop()])

        # One access disallowed for any group
        ModelAccess.write([model_access_wo_group], {
                'perm_delete': False,
                })
        self.assertRaises(UserError, TestAccess.delete, [tests.pop()])

        # Two access rules with one group allowed
        group = Group.search([('users', '=', Transaction().user)])[0]
        model_access_w_group, = ModelAccess.create([{
                    'model': model.id,
                    'group': group.id,
                    'perm_delete': True,
                    }])

        TestAccess.delete([tests.pop()])

        # Two access rules with both allowed
        ModelAccess.write([model_access_wo_group], {
                'perm_delete': True,
                })
        TestAccess.delete([tests.pop()])

        # Two access rules with any group allowed
        ModelAccess.write([model_access_w_group], {
                'perm_delete': False,
                })
        TestAccess.delete([tests.pop()])

        # Two access rules with both disallowed
        ModelAccess.write([model_access_wo_group], {
                'perm_delete': False,
                })
        self.assertRaises(UserError, TestAccess.delete, [tests.pop()])

        # One access disallowed for one group
        ModelAccess.delete([model_access_wo_group])
        self.assertRaises(UserError, TestAccess.delete, [tests.pop()])

        # One access allowed for one group
        ModelAccess.write([model_access_w_group], {
                'perm_delete': True,
                })
        TestAccess.delete([tests.pop()])

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        ModelAccess.write([model_access_w_group], {
                'group': group.id,
                })
        TestAccess.delete([tests.pop()])

        # One access disallowed for one other group
        ModelAccess.write([model_access_w_group], {
                'perm_delete': False,
                })
        TestAccess.delete([tests.pop()])


class ModelFieldAccessTestCase(unittest.TestCase):
    'Test Model Field Access'

    @classmethod
    def setUpClass(cls):
        activate_module('tests')

    @with_transaction(context=_context)
    def test0010perm_read(self):
        'Test Read Access'
        pool = Pool()
        FieldAccess = pool.get('ir.model.field.access')
        TestAccess = pool.get('test.access')
        Field = pool.get('ir.model.field')
        Group = pool.get('res.group')
        transaction = Transaction()

        field1, = Field.search([
                ('model.model', '=', 'test.access'),
                ('name', '=', 'field1'),
                ])
        field2, = Field.search([
                ('model.model', '=', 'test.access'),
                ('name', '=', 'field2'),
                ])

        test, = TestAccess.create([{
                    'field1': 'ham',
                    'field2': 'spam',
                    }])

        # Without field access
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # With field access

        # One access allowed for any group
        field_access_wo_group, = FieldAccess.create([{
                    'field': field1.id,
                    'group': None,
                    'perm_read': True,
                    }])
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # One access disallowed for any group
        FieldAccess.write([field_access_wo_group], {
                'perm_read': False,
                })

        self.assertRaises(UserError, TestAccess.read, [test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        self.assertRaises(UserError, TestAccess.read, [test.id])
        self.assertRaises(UserError, getattr, test, 'field1')
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules with one group allowed
        group = Group.search([('users', '=', transaction.user)])[0]
        field_access_w_group, = FieldAccess.create([{
                    'field': field1.id,
                    'group': group.id,
                    'perm_read': True,
                    }])

        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules with both allowed
        FieldAccess.write([field_access_wo_group], {
                'perm_read': True,
                })
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules with any group allowed
        FieldAccess.write([field_access_w_group], {
                'perm_read': False,
                })
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules with both disallowed
        FieldAccess.write([field_access_wo_group], {
            'perm_read': False,
            })
        self.assertRaises(UserError, TestAccess.read, [test.id],
            ['field1'])
        TestAccess.read([test.id], ['field2'])
        self.assertRaises(UserError, TestAccess.read, [test.id])
        self.assertRaises(UserError, getattr, test, 'field1')
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # One access disallowed for one group
        FieldAccess.delete([field_access_wo_group])
        self.assertRaises(UserError, TestAccess.read, [test.id],
            ['field1'])
        TestAccess.read([test.id], ['field2'])
        self.assertRaises(UserError, TestAccess.read, [test.id])
        self.assertRaises(UserError, getattr, test, 'field1')
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # One access allowed for one group
        FieldAccess.write([field_access_w_group], {
                'perm_read': True,
                })
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        FieldAccess.write([field_access_w_group], {
                'group': group.id,
                })
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # One access disallowed for one other group
        FieldAccess.write([field_access_w_group], {
                'perm_read': False,
                })
        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules on both fields allowed
        FieldAccess.delete([field_access_w_group])

        field_access1, = FieldAccess.create([{
                    'field': field1.id,
                    'group': None,
                    'perm_read': True,
                    }])
        field_access2, = FieldAccess.create([{
                    'field': field2.id,
                    'group': None,
                    'perm_read': True,
                    }])

        TestAccess.read([test.id], ['field1'])
        TestAccess.read([test.id], ['field2'])
        TestAccess.read([test.id])
        test.field1
        test.field2
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules on both fields one allowed and one disallowed
        FieldAccess.write([field_access2], {
            'perm_read': False,
            })
        TestAccess.read([test.id], ['field1'])
        self.assertRaises(UserError, TestAccess.read, [test.id],
            ['field2'])
        self.assertRaises(UserError, TestAccess.read, [test.id])
        test.field1
        self.assertRaises(UserError, getattr, test, 'field2')
        transaction.cache.clear()
        test = TestAccess(test.id)

        # Two access rules on both fields disallowed
        FieldAccess.write([field_access1], {
                'perm_read': False,
                })
        self.assertRaises(UserError, TestAccess.read, [test.id],
            ['field1'])
        self.assertRaises(UserError, TestAccess.read, [test.id],
            ['field2'])
        self.assertRaises(UserError, TestAccess.read, [test.id])
        self.assertRaises(UserError, getattr, test, 'field1')
        self.assertRaises(UserError, getattr, test, 'field2')
        transaction.cache.clear()
        test = TestAccess(test.id)

    @with_transaction(context=_context)
    def test_perm_write(self):
        'Test Write Access'
        pool = Pool()
        FieldAccess = pool.get('ir.model.field.access')
        TestAccess = pool.get('test.access')
        Field = pool.get('ir.model.field')
        Group = pool.get('res.group')
        transaction = Transaction()

        field1, = Field.search([
                ('model.model', '=', 'test.access'),
                ('name', '=', 'field1'),
                ])
        field2, = Field.search([
                ('model.model', '=', 'test.access'),
                ('name', '=', 'field2'),
                ])

        test, = TestAccess.create([{
                    'field1': 'ham',
                    'field2': 'spam',
                    }])

        # Without field access
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})

        # With field access

        # One access allowed for any group
        field_access_wo_group, = FieldAccess.create([{
                    'field': field1.id,
                    'group': None,
                    'perm_write': True,
                    }])
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # One access disallowed for any group
        FieldAccess.write([field_access_wo_group], {
                'perm_write': False,
                })

        TestAccess.write([test], {})
        self.assertRaises(UserError, TestAccess.write, [test],
            {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham',
                'field2': 'spam',
                })
        self.assertRaises(UserError, TestAccess.write,
            [test], {'field2': 'spam'}, [test], {'field1': 'ham'})

        # Two access rules with one group allowed
        group = Group.search([('users', '=', transaction.user)])[0]
        field_access_w_group, = FieldAccess.create([{
                    'field': field1.id,
                    'group': group.id,
                    'perm_write': True,
                    }])

        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules with both allowed
        FieldAccess.write([field_access_wo_group], {
                'perm_write': True,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules with any group allowed
        FieldAccess.write([field_access_w_group], {
                'perm_write': False,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules with both disallowed
        FieldAccess.write([field_access_wo_group], {
                'perm_write': False,
                })
        TestAccess.write([test], {})
        self.assertRaises(UserError, TestAccess.write, [test],
            {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # One access disallowed for one group
        FieldAccess.delete([field_access_wo_group])
        TestAccess.write([test], {})
        self.assertRaises(UserError, TestAccess.write, [test],
            {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'ham'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # One access allowed for one group
        FieldAccess.write([field_access_w_group], {
                'perm_write': True,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # One access allowed for one other group
        group, = Group.create([{'name': 'Test'}])
        FieldAccess.write([field_access_w_group], {
                'group': group.id,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # One access disallowed for one other group
        FieldAccess.write([field_access_w_group], {
                'perm_write': False,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules on both fields allowed
        FieldAccess.delete([field_access_w_group])

        field_access1, = FieldAccess.create([{
                    'field': field1.id,
                    'group': None,
                    'perm_write': True,
                    }])
        field_access2, = FieldAccess.create([{
                    'field': field2.id,
                    'group': None,
                    'perm_write': True,
                    }])

        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        TestAccess.write([test], {'field2': 'spam'})
        TestAccess.write([test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules on both fields one allowed and one disallowed
        FieldAccess.write([field_access2], {
                'perm_write': False,
                })
        TestAccess.write([test], {})
        TestAccess.write([test], {'field1': 'ham'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field2': 'spam'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham',
                'field2': 'spam',
                })

        # Two access rules on both fields disallowed
        FieldAccess.write([field_access1], {
                'perm_write': False,
                })
        TestAccess.write([test], {})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field2': 'spam'})
        self.assertRaises(UserError, TestAccess.write, [test], {
                'field1': 'ham',
                'field2': 'spam',
                })


def suite():
    suite_ = unittest.TestSuite()
    suite_.addTests(unittest.TestLoader(
        ).loadTestsFromTestCase(ModelAccessTestCase))
    suite_.addTests(unittest.TestLoader(
        ).loadTestsFromTestCase(ModelFieldAccessTestCase))
    return suite_



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from ..pool import Pool
from .configuration import *
from .translation import *
from .sequence import *
from .ui.menu import *
from .ui.view import *
from .ui.icon import *
from .action import *
from .model import *
from .attachment import *
from .note import *
from .cron import *
from .lang import *
from .export import *
from .rule import *
from .module import *
from .cache import *
from .date import *
from .trigger import *
from .session import *


def register():
    Pool.register(
        Configuration,
        Translation,
        TranslationSetStart,
        TranslationSetSucceed,
        TranslationCleanStart,
        TranslationCleanSucceed,
        TranslationUpdateStart,
        TranslationExportStart,
        TranslationExportResult,
        SequenceType,
        Sequence,
        SequenceStrict,
        UIMenu,
        UIMenuFavorite,
        View,
        ShowViewStart,
        ViewTreeWidth,
        ViewTreeState,
        ViewSearch,
        Icon,
        Action,
        ActionKeyword,
        ActionReport,
        ActionActWindow,
        ActionActWindowView,
        ActionActWindowDomain,
        ActionWizard,
        ActionURL,
        Model,
        ModelField,
        ModelAccess,
        ModelFieldAccess,
        ModelButton,
        ModelButtonRule,
        ModelButtonClick,
        ModelButtonReset,
        ModelData,
        PrintModelGraphStart,
        Attachment,
        Note,
        NoteRead,
        Cron,
        Lang,
        Export,
        ExportLine,
        RuleGroup,
        Rule,
        Module,
        ModuleDependency,
        ModuleConfigWizardItem,
        ModuleConfigWizardFirst,
        ModuleConfigWizardOther,
        ModuleConfigWizardDone,
        ModuleActivateUpgradeStart,
        ModuleActivateUpgradeDone,
        Cache,
        Date,
        Trigger,
        TriggerLog,
        Session,
        SessionWizard,
        module='ir', type_='model')
    Pool.register(
        TranslationSet,
        TranslationClean,
        TranslationUpdate,
        TranslationExport,
        TranslationReport,
        ShowView,
        PrintModelGraph,
        ModuleConfigWizard,
        ModuleActivateUpgrade,
        ModuleConfig,
        module='ir', type_='wizard')
    Pool.register(
        ModelGraph,
        ModelWorkflowGraph,
        module='ir', type_='report')



# -*- coding: utf-8 -*-
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import unittest
import time
import datetime
from itertools import combinations

from trytond.tests.test_tryton import activate_module, with_transaction
from trytond.tests.trigger import TRIGGER_LOGS
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.exceptions import UserError
from trytond.pyson import PYSONEncoder, Eval


class TriggerTestCase(unittest.TestCase):
    'Test Trigger'

    @classmethod
    def setUpClass(cls):
        activate_module('tests')

    @with_transaction()
    def test_constraints(self):
        'Test constraints'
        pool = Pool()
        Model = pool.get('ir.model')
        Trigger = pool.get('ir.trigger')
        transaction = Transaction()

        model, = Model.search([
                ('model', '=', 'test.triggered'),
                ])
        action_model, = Model.search([
                ('model', '=', 'test.trigger_action'),
                ])

        values = {
            'name': 'Test',
            'model': model.id,
            'on_time': True,
            'condition': 'true',
            'action_model': action_model.id,
            'action_function': 'test',
            }
        self.assert_(Trigger.create([values]))

        transaction.rollback()

        # on_exclusive
        for i in range(1, 4):
            for combination in combinations(
                    ['create', 'write', 'delete'], i):
                combination_values = values.copy()
                for mode in combination:
                    combination_values['on_%s' % mode] = True
                self.assertRaises(UserError, Trigger.create,
                    [combination_values])
                transaction.rollback()

        # check_condition
        condition_values = values.copy()
        condition_values['condition'] = '='
        self.assertRaises(UserError, Trigger.create,
            [condition_values])
        transaction.rollback()

        # Restart the cache on the get_triggers method of ir.trigger
        Trigger._get_triggers_cache.clear()

    @with_transaction()
    def test_on_create(self):
        'Test on_create'
        pool = Pool()
        Model = pool.get('ir.model')
        Trigger = pool.get('ir.trigger')
        Triggered = pool.get('test.triggered')

        model, = Model.search([
                ('model', '=', 'test.triggered'),
                ])
        action_model, = Model.search([
                ('model', '=', 'test.trigger_action'),
                ])

        trigger, = Trigger.create([{
                    'name': 'Test',
                    'model': model.id,
                    'on_create': True,
                    'condition': 'true',
                    'action_model': action_model.id,
                    'action_function': 'trigger',
                    }])

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])

        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Trigger with condition
        condition = PYSONEncoder().encode(
            Eval('self', {}).get('name') == 'Bar')
        Trigger.write([trigger], {
                'condition': condition,
                })

        # Matching condition
        triggered, = Triggered.create([{
                    'name': 'Bar',
                    }])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Non matching condition
        triggered, = Triggered.create([{
                    'name': 'Foo',
                    }])
        self.assertEqual(TRIGGER_LOGS, [])

        # With limit number
        Trigger.write([trigger], {
                'condition': 'true',
                'limit_number': 1,
                })
        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # With minimum delay
        Trigger.write([trigger], {
                'limit_number': 0,
                'minimum_time_delay': datetime.timedelta(hours=1),
                })
        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Restart the cache on the get_triggers method of ir.trigger
        Trigger._get_triggers_cache.clear()

    @with_transaction()
    def test0030on_write(self):
        'Test on_write'
        pool = Pool()
        Model = pool.get('ir.model')
        Trigger = pool.get('ir.trigger')
        Triggered = pool.get('test.triggered')

        model, = Model.search([
                ('model', '=', 'test.triggered'),
                ])
        action_model, = Model.search([
                ('model', '=', 'test.trigger_action'),
                ])

        trigger, = Trigger.create([{
                    'name': 'Test',
                    'model': model.id,
                    'on_write': True,
                    'condition': 'true',
                    'action_model': action_model.id,
                    'action_function': 'trigger',
                    }])

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])

        Triggered.write([triggered], {
                'name': 'Foo',
                })
        self.assertEqual(TRIGGER_LOGS, [])

        # Trigger with condition
        condition = PYSONEncoder().encode(
            Eval('self', {}).get('name') == 'Bar')
        Trigger.write([trigger], {
                'condition': condition,
                })

        # Matching condition
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # No change in condition
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        self.assertEqual(TRIGGER_LOGS, [])

        # Different change in condition
        Triggered.write([triggered], {
                'name': 'Foo',
                })
        self.assertEqual(TRIGGER_LOGS, [])

        # With limit number
        condition = PYSONEncoder().encode(
            Eval('self', {}).get('name') == 'Bar')
        Trigger.write([trigger], {
                'condition': condition,
                'limit_number': 1,
                })
        triggered, = Triggered.create([{
                    'name': 'Foo',
                    }])
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        Triggered.write([triggered], {
                'name': 'Foo',
                })
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # With minimum delay
        Trigger.write([trigger], {
                'limit_number': 0,
                'minimum_time_delay': datetime.timedelta.max,
                })
        triggered, = Triggered.create([{
                    'name': 'Foo',
                    }])
        for name in ('Bar', 'Foo', 'Bar'):
            Triggered.write([triggered], {
                    'name': name,
                    })
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        Trigger.write([trigger], {
                'minimum_time_delay': datetime.timedelta(seconds=1),
                })
        triggered, = Triggered.create([{
                    'name': 'Foo',
                    }])
        for name in ('Bar', 'Foo'):
            Triggered.write([triggered], {
                    'name': name,
                    })
        time.sleep(1.2)
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        self.assertEqual(TRIGGER_LOGS,
            [([triggered], trigger), ([triggered], trigger)])
        TRIGGER_LOGS.pop()
        TRIGGER_LOGS.pop()

        # Restart the cache on the get_triggers method of ir.trigger
        Trigger._get_triggers_cache.clear()

    @with_transaction()
    def test0040on_delete(self):
        'Test on_delete'
        pool = Pool()
        Model = pool.get('ir.model')
        Trigger = pool.get('ir.trigger')
        Triggered = pool.get('test.triggered')
        TriggerLog = pool.get('ir.trigger.log')

        model, = Model.search([
                ('model', '=', 'test.triggered'),
                ])
        action_model, = Model.search([
                ('model', '=', 'test.trigger_action'),
                ])

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])

        trigger, = Trigger.create([{
                    'name': 'Test',
                    'model': model.id,
                    'on_delete': True,
                    'condition': 'true',
                    'action_model': action_model.id,
                    'action_function': 'trigger',
                    }])

        Triggered.delete([triggered])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()
        Transaction().delete = {}

        # Trigger with condition
        condition = PYSONEncoder().encode(
            Eval('self', {}).get('name') == 'Bar')
        Trigger.write([trigger], {
                'condition': condition,
                })

        triggered, = Triggered.create([{
                    'name': 'Bar',
                    }])

        # Matching condition
        Triggered.delete([triggered])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()
        Transaction().delete = {}

        triggered, = Triggered.create([{
                    'name': 'Foo',
                    }])

        # Non matching condition
        Triggered.delete([triggered])
        self.assertEqual(TRIGGER_LOGS, [])
        Transaction().delete = {}

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])

        # With limit number
        Trigger.write([trigger], {
                'condition': 'true',
                'limit_number': 1,
                })
        Triggered.delete([triggered])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()
        Transaction().delete = {}
        # Delete trigger logs because SQLite reuse the same triggered_id
        TriggerLog.delete(TriggerLog.search([
                    ('trigger', '=', trigger.id),
                    ]))

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])

        # With minimum delay
        Trigger.write([trigger], {
                'limit_number': 0,
                'minimum_time_delay': datetime.timedelta(hours=1),
                })
        Triggered.delete([triggered])
        self.assertEqual(TRIGGER_LOGS, [([triggered], trigger)])
        TRIGGER_LOGS.pop()
        Transaction().delete = {}

        # Restart the cache on the get_triggers method of ir.trigger
        Trigger._get_triggers_cache.clear()

    @with_transaction()
    def test_on_time(self):
        'Test on_time'
        pool = Pool()
        Model = pool.get('ir.model')
        Trigger = pool.get('ir.trigger')
        Triggered = pool.get('test.triggered')
        TriggerLog = pool.get('ir.trigger.log')

        model, = Model.search([
                ('model', '=', 'test.triggered'),
                ])
        action_model, = Model.search([
                ('model', '=', 'test.trigger_action'),
                ])

        trigger, = Trigger.create([{
                    'name': 'Test',
                    'model': model.id,
                    'on_time': True,
                    'condition': 'true',
                    'action_model': action_model.id,
                    'action_function': 'trigger',
                    }])

        triggered, = Triggered.create([{
                    'name': 'Test',
                    }])
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Trigger with condition
        condition = PYSONEncoder().encode(
            Eval('self', {}).get('name') == 'Bar')
        Trigger.write([trigger], {
                'condition': condition,
                })

        # Matching condition
        Triggered.write([triggered], {
                'name': 'Bar',
                })
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Non matching condition
        Triggered.write([triggered], {
                'name': 'Foo',
                })
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [])

        # With limit number
        Trigger.write([trigger], {
                'condition': 'true',
                'limit_number': 1,
                })
        Trigger.trigger_time()
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [([triggered], trigger)])
        TRIGGER_LOGS.pop()

        # Delete trigger logs of limit number test
        TriggerLog.delete(TriggerLog.search([
                    ('trigger', '=', trigger.id),
                    ]))

        # With minimum delay
        Trigger.write([trigger], {
                'limit_number': 0,
                'minimum_time_delay': datetime.timedelta.max,
                })
        Trigger.trigger_time()
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [([triggered], trigger)])
        TRIGGER_LOGS.pop()
        Transaction().delete = {}

        # Delete trigger logs of previous minimum delay test
        TriggerLog.delete(TriggerLog.search([
                    ('trigger', '=', trigger.id),
                    ]))

        Trigger.write([trigger], {
                'minimum_time_delay': datetime.timedelta(seconds=1),
                })
        Trigger.trigger_time()
        time.sleep(1.2)
        Trigger.trigger_time()
        self.assert_(TRIGGER_LOGS == [([triggered], trigger),
                ([triggered], trigger)])
        TRIGGER_LOGS.pop()
        TRIGGER_LOGS.pop()
        Transaction().delete = {}

        # Restart the cache on the get_triggers method of ir.trigger
        Trigger._get_triggers_cache.clear()


def suite():
    return unittest.TestLoader().loadTestsFromTestCase(TriggerTestCase)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import datetime
from decimal import Decimal
from trytond.model import (
    ModelSQL, ModelStorage, DictSchemaMixin, fields, Unique)
from trytond.pyson import Eval

__all__ = [
    'Boolean', 'BooleanDefault',
    'Integer', 'IntegerDefault', 'IntegerRequired', 'IntegerDomain',
    'Float', 'FloatDefault', 'FloatRequired', 'FloatDigits',
    'Numeric', 'NumericDefault', 'NumericRequired', 'NumericDigits',
    'Char', 'CharDefault', 'CharRequired', 'CharSize', 'CharTranslate',
    'Text', 'TextDefault', 'TextRequired', 'TextSize', 'TextTranslate',
    'Date', 'DateDefault', 'DateRequired',
    'DateTime', 'DateTimeDefault', 'DateTimeRequired', 'DateTimeFormat',
    'Time', 'TimeDefault', 'TimeRequired', 'TimeFormat',
    'TimeDelta', 'TimeDeltaDefault', 'TimeDeltaRequired',
    'One2One', 'One2OneTarget', 'One2OneRelation', 'One2OneRequired',
    'One2OneRequiredRelation', 'One2OneDomain', 'One2OneDomainRelation',
    'One2Many', 'One2ManyTarget',
    'One2ManyRequired', 'One2ManyRequiredTarget',
    'One2ManyReference', 'One2ManyReferenceTarget',
    'One2ManySize', 'One2ManySizeTarget',
    'One2ManySizePYSON', 'One2ManySizePYSONTarget',
    'One2ManyFilter', 'One2ManyFilterTarget',
    'One2ManyFilterDomain', 'One2ManyFilterDomainTarget',
    'Many2Many', 'Many2ManyTarget', 'Many2ManyRelation',
    'Many2ManyRequired', 'Many2ManyRequiredTarget',
    'Many2ManyRequiredRelation',
    'Many2ManyReference', 'Many2ManyReferenceTarget',
    'Many2ManyReferenceRelation',
    'Many2ManySize', 'Many2ManySizeTarget', 'Many2ManySizeRelation',
    'Many2ManyTree', 'Many2ManyTreeRelation',
    'Many2ManyFilter', 'Many2ManyFilterTarget', 'Many2ManyFilterRelation',
    'Many2ManyFilterDomain', 'Many2ManyFilterDomainTarget',
    'Many2ManyFilterDomainRelation',
    'Reference', 'ReferenceTarget', 'ReferenceRequired',
    'Selection', 'SelectionRequired',
    'DictSchema', 'Dict', 'DictDefault', 'DictRequired', 'DictJSONB',
    'Binary', 'BinaryDefault', 'BinaryRequired', 'BinaryFileStorage',
    'Many2OneDomainValidation', 'Many2OneTarget', 'Many2OneOrderBy',
    'Many2OneSearch', 'Many2OneTree', 'Many2OneMPTT', 'Many2OneNoForeignKey',
    'Many2OneTargetStorage'
    ]


class Boolean(ModelSQL):
    'Boolean'
    __name__ = 'test.boolean'
    boolean = fields.Boolean(string='Boolean', help='Test boolean')


class BooleanDefault(ModelSQL):
    'Boolean Default'
    __name__ = 'test.boolean_default'
    boolean = fields.Boolean(string='Boolean', help='Test boolean')

    @staticmethod
    def default_boolean():
        return True


class Integer(ModelSQL):
    'Integer'
    __name__ = 'test.integer'
    integer = fields.Integer(string='Integer', help='Test integer',
            required=False)


class IntegerDefault(ModelSQL):
    'Integer Default'
    __name__ = 'test.integer_default'
    integer = fields.Integer(string='Integer', help='Test integer',
            required=False)

    @staticmethod
    def default_integer():
        return 5


class IntegerRequired(ModelSQL):
    'Integer Required'
    __name__ = 'test.integer_required'
    integer = fields.Integer(string='Integer', help='Test integer',
            required=True)


class IntegerDomain(ModelSQL):
    'Integer Domain'
    __name__ = 'test.integer_domain'
    integer = fields.Integer('Integer', domain=[('integer', '>', 42)])


class Float(ModelSQL):
    'Float'
    __name__ = 'test.float'
    float = fields.Float(string='Float', help='Test float',
            required=False)


class FloatDefault(ModelSQL):
    'Float Default'
    __name__ = 'test.float_default'
    float = fields.Float(string='Float', help='Test float',
            required=False)

    @staticmethod
    def default_float():
        return 5.5


class FloatRequired(ModelSQL):
    'Float Required'
    __name__ = 'test.float_required'
    float = fields.Float(string='Float', help='Test float',
            required=True)


class FloatDigits(ModelSQL):
    'Float Digits'
    __name__ = 'test.float_digits'
    digits = fields.Integer('Digits')
    float = fields.Float(string='Float', help='Test float',
        required=False, digits=(16, Eval('digits', 2)),
        depends=['digits'])


class Numeric(ModelSQL):
    'Numeric'
    __name__ = 'test.numeric'
    numeric = fields.Numeric(string='Numeric', help='Test numeric',
            required=False)


class NumericDefault(ModelSQL):
    'Numeric Default'
    __name__ = 'test.numeric_default'
    numeric = fields.Numeric(string='Numeric', help='Test numeric',
            required=False)

    @staticmethod
    def default_numeric():
        return Decimal('5.5')


class NumericRequired(ModelSQL):
    'Numeric Required'
    __name__ = 'test.numeric_required'
    numeric = fields.Numeric(string='Numeric', help='Test numeric',
            required=True)


class NumericDigits(ModelSQL):
    'Numeric Digits'
    __name__ = 'test.numeric_digits'
    digits = fields.Integer('Digits')
    numeric = fields.Numeric(string='Numeric', help='Test numeric',
        required=False, digits=(16, Eval('digits', 2)),
        depends=['digits'])


class Char(ModelSQL):
    'Char'
    __name__ = 'test.char'
    char = fields.Char(string='Char', size=None, help='Test char',
            required=False)


class CharDefault(ModelSQL):
    'Char Default'
    __name__ = 'test.char_default'
    char = fields.Char(string='Char', size=None, help='Test char',
            required=False)

    @staticmethod
    def default_char():
        return 'Test'


class CharRequired(ModelSQL):
    'Char Required'
    __name__ = 'test.char_required'
    char = fields.Char(string='Char', size=None, help='Test char',
            required=True)


class CharSize(ModelSQL):
    'Char Size'
    __name__ = 'test.char_size'
    char = fields.Char(string='Char', size=5, help='Test char',
            required=False)


class CharTranslate(ModelSQL):
    'Char Translate'
    __name__ = 'test.char_translate'
    char = fields.Char(string='Char', size=None, help='Test char',
            required=False, translate=True)


class Text(ModelSQL):
    'Text'
    __name__ = 'test.text'
    text = fields.Text(string='Text', size=None, help='Test text',
            required=False)


class TextDefault(ModelSQL):
    'Text Default'
    __name__ = 'test.text_default'
    text = fields.Text(string='Text', size=None, help='Test text',
            required=False)

    @staticmethod
    def default_text():
        return 'Test'


class TextRequired(ModelSQL):
    'Text Required'
    __name__ = 'test.text_required'
    text = fields.Text(string='Text', size=None, help='Test text',
            required=True)


class TextSize(ModelSQL):
    'Text Size'
    __name__ = 'test.text_size'
    text = fields.Text(string='Text', size=5, help='Test text',
            required=False)


class TextTranslate(ModelSQL):
    'Text Translate'
    __name__ = 'test.text_translate'
    text = fields.Text(string='Text', size=None, help='Test text',
            required=False, translate=True)


class Date(ModelSQL):
    'Date'
    __name__ = 'test.date'
    date = fields.Date(string='Date', help='Test date',
            required=False)


class DateDefault(ModelSQL):
    'Date Default'
    __name__ = 'test.date_default'
    date = fields.Date(string='Date', help='Test date',
            required=False)

    @staticmethod
    def default_date():
        return datetime.date(2000, 1, 1)


class DateRequired(ModelSQL):
    'Date Required'
    __name__ = 'test.date_required'
    date = fields.Date(string='Date', help='Test date',
            required=True)


class DateTime(ModelSQL):
    'DateTime'
    __name__ = 'test.datetime'
    datetime = fields.DateTime(string='DateTime', help='Test datetime',
            required=False)


class DateTimeDefault(ModelSQL):
    'DateTime Default'
    __name__ = 'test.datetime_default'
    datetime = fields.DateTime(string='DateTime', help='Test datetime',
            required=False)

    @staticmethod
    def default_datetime():
        return datetime.datetime(2000, 1, 1, 12, 0, 0, 0)


class DateTimeRequired(ModelSQL):
    'DateTime Required'
    __name__ = 'test.datetime_required'
    datetime = fields.DateTime(string='DateTime', help='Test datetime',
            required=True)


class DateTimeFormat(ModelSQL):
    'DateTime Format'
    __name__ = 'test.datetime_format'
    datetime = fields.DateTime(string='DateTime', format='%H:%M')


class Time(ModelSQL):
    'Time'
    __name__ = 'test.time'
    time = fields.Time(string='Time', help='Test time', required=False)


class TimeDefault(ModelSQL):
    'Time Default'
    __name__ = 'test.time_default'
    time = fields.Time(string='Time', help='Test time', required=False)

    @staticmethod
    def default_time():
        return datetime.time(16, 30)


class TimeRequired(ModelSQL):
    'Time'
    __name__ = 'test.time_required'
    time = fields.Time(string='Time', help='Test time', required=True)


class TimeFormat(ModelSQL):
    'Time Format'
    __name__ = 'test.time_format'
    time = fields.Time(string='Time', format='%H:%M')


class TimeDelta(ModelSQL):
    'TimeDelta'
    __name__ = 'test.timedelta'
    timedelta = fields.TimeDelta(string='TimeDelta', help='Test timedelta',
        required=False)


class TimeDeltaDefault(ModelSQL):
    'TimeDelta Default'
    __name__ = 'test.timedelta_default'
    timedelta = fields.TimeDelta(string='TimeDelta', help='Test timedelta',
        required=False)

    @staticmethod
    def default_timedelta():
        return datetime.timedelta(seconds=3600)


class TimeDeltaRequired(ModelSQL):
    'TimeDelta Required'
    __name__ = 'test.timedelta_required'
    timedelta = fields.TimeDelta(string='TimeDelta', help='Test timedelta',
        required=True)


class One2One(ModelSQL):
    'One2One'
    __name__ = 'test.one2one'
    name = fields.Char('Name', required=True)
    one2one = fields.One2One('test.one2one.relation', 'origin', 'target',
            string='One2One', help='Test one2one', required=False)


class One2OneTarget(ModelSQL):
    'One2One Target'
    __name__ = 'test.one2one.target'
    name = fields.Char('Name', required=True)


class One2OneRelation(ModelSQL):
    'One2One Relation'
    __name__ = 'test.one2one.relation'
    origin = fields.Many2One('test.one2one', 'Origin')
    target = fields.Many2One('test.one2one.target', 'Target')

    @classmethod
    def __setup__(cls):
        super(One2OneRelation, cls).__setup__()
        table = cls.__table__()
        cls._sql_constraints += [
            ('origin_unique', Unique(table, table.origin),
                'Origin must be unique'),
            ('target_unique', Unique(table, table.target),
                'Target must be unique'),
            ]


class One2OneRequired(ModelSQL):
    'One2One'
    __name__ = 'test.one2one_required'
    name = fields.Char('Name', required=True)
    one2one = fields.One2One('test.one2one_required.relation', 'origin',
        'target', string='One2One', help='Test one2one', required=True)


class One2OneRequiredRelation(ModelSQL):
    'One2One Relation'
    __name__ = 'test.one2one_required.relation'
    origin = fields.Many2One('test.one2one_required', 'Origin')
    target = fields.Many2One('test.one2one.target', 'Target')

    @classmethod
    def __setup__(cls):
        super(One2OneRequiredRelation, cls).__setup__()
        table = cls.__table__()
        cls._sql_constraints += [
            ('origin_unique', Unique(table, table.origin),
                'Origin must be unique'),
            ('target_unique', Unique(table, table.target),
                'Target must be unique'),
            ]


class One2OneDomain(ModelSQL):
    'One2One'
    __name__ = 'test.one2one_domain'
    name = fields.Char('Name', required=True)
    one2one = fields.One2One('test.one2one_domain.relation', 'origin',
        'target', string='One2One', help='Test one2one',
        domain=[('name', '=', 'domain')])


class One2OneDomainRelation(ModelSQL):
    'One2One Relation'
    __name__ = 'test.one2one_domain.relation'
    origin = fields.Many2One('test.one2one_domain', 'Origin')
    target = fields.Many2One('test.one2one.target', 'Target')

    @classmethod
    def __setup__(cls):
        super(One2OneDomainRelation, cls).__setup__()
        table = cls.__table__()
        cls._sql_constraints += [
            ('origin_unique', Unique(table, table.origin),
                'Origin must be unique'),
            ('target_unique', Unique(table, table.target),
                'Target must be unique'),
            ]


class One2Many(ModelSQL):
    'One2Many Relation'
    __name__ = 'test.one2many'
    name = fields.Char('Name', required=True)
    targets = fields.One2Many('test.one2many.target', 'origin', 'Targets')


class One2ManyTarget(ModelSQL):
    'One2Many Target'
    __name__ = 'test.one2many.target'
    name = fields.Char('Name', required=True)
    origin = fields.Many2One('test.one2many', 'Origin')


class One2ManyRequired(ModelSQL):
    'One2Many Required'
    __name__ = 'test.one2many_required'
    name = fields.Char('Name', required=True)
    targets = fields.One2Many('test.one2many_required.target', 'origin',
        'Targets', required=True)


class One2ManyRequiredTarget(ModelSQL):
    'One2Many Required Target'
    __name__ = 'test.one2many_required.target'
    name = fields.Char('Name', required=True)
    origin = fields.Many2One('test.one2many_required', 'Origin')


class One2ManyReference(ModelSQL):
    'One2Many Reference Relation'
    __name__ = 'test.one2many_reference'
    name = fields.Char('Name', required=True)
    targets = fields.One2Many('test.one2many_reference.target', 'origin',
        'Targets')


class One2ManyReferenceTarget(ModelSQL):
    'One2Many Reference Target'
    __name__ = 'test.one2many_reference.target'
    name = fields.Char('Name', required=True)
    origin = fields.Reference('Origin', [
            (None, ''),
            ('test.one2many_reference', 'One2Many Reference'),
            ])


class One2ManySize(ModelSQL):
    'One2Many Size Relation'
    __name__ = 'test.one2many_size'
    targets = fields.One2Many('test.one2many_size.target', 'origin', 'Targets',
        size=3)


class One2ManySizeTarget(ModelSQL):
    'One2Many Size Target'
    __name__ = 'test.one2many_size.target'
    origin = fields.Many2One('test.one2many_size', 'Origin')


class One2ManySizePYSON(ModelSQL):
    'One2Many Size PYSON Relation'
    __name__ = 'test.one2many_size_pyson'
    limit = fields.Integer('Limit')
    targets = fields.One2Many('test.one2many_size_pyson.target', 'origin',
        'Targets', size=Eval('limit', 0))


class One2ManySizePYSONTarget(ModelSQL):
    'One2Many Size PYSON Target'
    __name__ = 'test.one2many_size_pyson.target'
    origin = fields.Many2One('test.one2many_size_pyson', 'Origin')


class One2ManyFilter(ModelSQL):
    'One2Many Filter Relation'
    __name__ = 'test.one2many_filter'
    targets = fields.One2Many('test.one2many_filter.target', 'origin',
        'Targets')
    filtered_targets = fields.One2Many('test.one2many_filter.target', 'origin',
        'Filtered Targets', filter=[('value', '>', 2)])


class One2ManyFilterTarget(ModelSQL):
    'One2Many Filter Target'
    __name__ = 'test.one2many_filter.target'
    origin = fields.Many2One('test.one2many_filter', 'Origin')
    value = fields.Integer('Value')


class One2ManyFilterDomain(ModelSQL):
    'One2Many Filter Relation'
    __name__ = 'test.one2many_filter_domain'
    targets = fields.One2Many('test.one2many_filter_domain.target', 'origin',
        'Targets', domain=[('value', '<', 10)])
    filtered_targets = fields.One2Many('test.one2many_filter_domain.target',
        'origin', 'Filtered Targets', domain=[('value', '<', 10)],
        filter=[('value', '>', 2)])


class One2ManyFilterDomainTarget(ModelSQL):
    'One2Many Filter Domain Target'
    __name__ = 'test.one2many_filter_domain.target'
    origin = fields.Many2One('test.one2many_filter_domain', 'Origin')
    value = fields.Integer('Value')


class Many2Many(ModelSQL):
    'Many2Many'
    __name__ = 'test.many2many'
    name = fields.Char('Name', required=True)
    targets = fields.Many2Many('test.many2many.relation', 'origin', 'target',
        'Targets')


class Many2ManyTarget(ModelSQL):
    'Many2Many Target'
    __name__ = 'test.many2many.target'
    name = fields.Char('Name', required=True)


class Many2ManyRelation(ModelSQL):
    'Many2Many Relation'
    __name__ = 'test.many2many.relation'
    origin = fields.Many2One('test.many2many', 'Origin')
    target = fields.Many2One('test.many2many.target', 'Target')


class Many2ManyRequired(ModelSQL):
    'Many2Many Required'
    __name__ = 'test.many2many_required'
    name = fields.Char('Name', required=True)
    targets = fields.Many2Many('test.many2many_required.relation', 'origin',
        'target', 'Targets', required=True)


class Many2ManyRequiredTarget(ModelSQL):
    'Many2Many Required Target'
    __name__ = 'test.many2many_required.target'
    name = fields.Char('Name', required=True)


class Many2ManyRequiredRelation(ModelSQL):
    'Many2Many Required Relation'
    __name__ = 'test.many2many_required.relation'
    origin = fields.Many2One('test.many2many_required', 'Origin')
    target = fields.Many2One('test.many2many_required.target', 'Target')


class Many2ManyReference(ModelSQL):
    'Many2Many Reference'
    __name__ = 'test.many2many_reference'
    name = fields.Char('Name', required=True)
    targets = fields.Many2Many('test.many2many_reference.relation', 'origin',
        'target', 'Targets')


class Many2ManyReferenceTarget(ModelSQL):
    'Many2Many Reference Target'
    __name__ = 'test.many2many_reference.target'
    name = fields.Char('Name', required=True)


class Many2ManyReferenceRelation(ModelSQL):
    'Many2Many Relation'
    __name__ = 'test.many2many_reference.relation'
    origin = fields.Reference('Origin', [
            (None, ''),
            ('test.many2many_reference', 'Many2Many Reference'),
            ])
    target = fields.Many2One('test.many2many_reference.target',
        'Reference Target')


class Many2ManySize(ModelSQL):
    'Many2Many Size Relation'
    __name__ = 'test.many2many_size'
    targets = fields.Many2Many('test.many2many_size.relation', 'origin',
        'target', 'Targets', size=5)


class Many2ManySizeTarget(ModelSQL):
    'Many2Many Size Target'
    __name__ = 'test.many2many_size.target'
    name = fields.Char('Name')


class Many2ManySizeRelation(ModelSQL):
    'Many2Many Size Relation'
    __name__ = 'test.many2many_size.relation'
    origin = fields.Many2One('test.many2many_size', 'Origin')
    target = fields.Many2One('test.many2many_size.target', 'Target')


class Many2ManyTree(ModelSQL):
    'Many2Many Tree'
    __name__ = 'test.many2many_tree'
    parents = fields.Many2Many('test.many2many_tree.relation',
        'child', 'parent', 'Parents')
    children = fields.Many2Many('test.many2many_tree.relation',
        'parent', 'child', 'Children')


class Many2ManyTreeRelation(ModelSQL):
    'Many2Many Tree Relation'
    __name__ = 'test.many2many_tree.relation'
    parent = fields.Many2One('test.many2many_tree', 'Parent')
    child = fields.Many2One('test.many2many_tree', 'Child')


class Many2ManyFilter(ModelSQL):
    'Many2Many Filter Relation'
    __name__ = 'test.many2many_filter'
    targets = fields.Many2Many('test.many2many_filter.relation', 'origin',
        'target', 'Targets')
    filtered_targets = fields.Many2Many('test.many2many_filter.relation',
        'origin', 'target', 'Targets',
        filter=[('value', '>', 2)])
    or_filtered_targets = fields.Many2Many('test.many2many_filter.relation',
        'origin', 'target', 'Targets',
        filter=['OR', ('value', '>', 2), ('value', '<', 0)])


class Many2ManyFilterTarget(ModelSQL):
    'Many2Many Filter Target'
    __name__ = 'test.many2many_filter.target'
    value = fields.Integer('Value')


class Many2ManyFilterRelation(ModelSQL):
    'Many2Many Filter Relation'
    __name__ = 'test.many2many_filter.relation'
    origin = fields.Many2One('test.many2many_filter', 'Origin')
    target = fields.Many2One('test.many2many_filter.target', 'Target')


class Many2ManyFilterDomain(ModelSQL):
    'Many2Many Filter Domain Relation'
    __name__ = 'test.many2many_filter_domain'
    targets = fields.Many2Many('test.many2many_filter_domain.relation',
        'origin', 'target', 'Targets', domain=[('value', '<', 10)])
    filtered_targets = fields.Many2Many(
        'test.many2many_filter_domain.relation', 'origin', 'target', 'Targets',
        domain=[('value', '<', 10)], filter=[('value', '>', 2)])


class Many2ManyFilterDomainTarget(ModelSQL):
    'Many2Many Filter Domain Target'
    __name__ = 'test.many2many_filter_domain.target'
    value = fields.Integer('Value')


class Many2ManyFilterDomainRelation(ModelSQL):
    'Many2Many Filter Domain Relation'
    __name__ = 'test.many2many_filter_domain.relation'
    origin = fields.Many2One('test.many2many_filter_domain', 'Origin')
    target = fields.Many2One('test.many2many_filter.target', 'Target')


class Reference(ModelSQL):
    'Reference'
    __name__ = 'test.reference'
    name = fields.Char('Name', required=True)
    reference = fields.Reference('Reference', selection=[
            (None, ''),
            ('test.reference.target', 'Target'),
            ])


class ReferenceTarget(ModelSQL):
    'Reference Target'
    __name__ = 'test.reference.target'
    name = fields.Char('Name', required=True)


class ReferenceRequired(ModelSQL):
    'Reference Required'
    __name__ = 'test.reference_required'
    name = fields.Char('Name', required=True)
    reference = fields.Reference('Reference', selection=[
            (None, ''),
            ('test.reference.target', 'Target'),
            ], required=True)


class Selection(ModelSQL):
    'Selection'
    __name__ = 'test.selection'
    select = fields.Selection([
            ('', ''), ('arabic', 'Arabic'), ('hexa', 'Hexadecimal')],
        'Selection')
    select_string = select.translated('select')
    dyn_select = fields.Selection('get_selection',
        'Instance Dynamic Selection')
    dyn_select_string = dyn_select.translated('dyn_select')
    dyn_select_static = fields.Selection('static_selection',
        'Static Selection')
    dyn_select_static_string = dyn_select_static.translated(
        'dyn_select_static')

    @fields.depends('select')
    def get_selection(self):
        if self.select == 'arabic':
            return [('', '')] + [(str(i), str(i)) for i in range(1, 11)]
        else:
            return [('', '')] + [(hex(i), hex(i)) for i in range(1, 11)]

    @staticmethod
    def static_selection():
        return [('', '')] + [(str(i), str(i)) for i in range(1, 11)]


class SelectionRequired(ModelSQL):
    'Selection Required'
    __name__ = 'test.selection_required'
    select = fields.Selection([('arabic', 'Arabic'), ('latin', 'Latin')],
        'Selection', required=True)


class DictSchema(DictSchemaMixin, ModelSQL):
    'Dict Schema'
    __name__ = 'test.dict.schema'


class Dict(ModelSQL):
    'Dict'
    __name__ = 'test.dict'
    dico = fields.Dict('test.dict.schema', 'Test Dict')
    dico_string = dico.translated('dico')
    dico_string_keys = dico.translated('dico', 'keys')


class DictDefault(ModelSQL):
    'Dict Default'
    __name__ = 'test.dict_default'
    dico = fields.Dict(None, 'Test Dict')

    @staticmethod
    def default_dico():
        return dict(a=1)


class DictRequired(ModelSQL):
    'Dict Required'
    __name__ = 'test.dict_required'
    dico = fields.Dict(None, 'Test Dict', required=True)


class DictJSONB(ModelSQL):
    'Dict JSONB'
    __name__ = 'test.dict_jsonb'
    dico = fields.Dict('test.dict.schema', 'Test Dict')


class Binary(ModelSQL):
    'Binary'
    __name__ = 'test.binary'
    binary = fields.Binary('Binary')


class BinaryDefault(ModelSQL):
    'Binary Default'
    __name__ = 'test.binary_default'
    binary = fields.Binary('Binary Default')

    @staticmethod
    def default_binary():
        return b'default'


class BinaryRequired(ModelSQL):
    'Binary Required'
    __name__ = 'test.binary_required'
    binary = fields.Binary('Binary Required', required=True)


class BinaryFileStorage(ModelSQL):
    "Binary in FileStorage"
    __name__ = 'test.binary_filestorage'
    binary = fields.Binary('Binary', file_id='binary_id')
    binary_id = fields.Char('Binary ID')


class Many2OneTarget(ModelSQL):
    "Many2One Domain Validation Target"
    __name__ = 'test.many2one_target'
    _order_name = 'value'

    active = fields.Boolean('Active')
    value = fields.Integer('Value')

    @staticmethod
    def default_active():
        return True


class Many2OneDomainValidation(ModelSQL):
    "Many2One Domain Validation"
    __name__ = 'test.many2one_domainvalidation'
    many2one = fields.Many2One('test.many2one_target',
        'many2one',
        domain=[
            ('value', '>', 5),
            ])
    dummy = fields.Char('Dummy')


class Many2OneOrderBy(ModelSQL):
    "Many2One OrderBy"
    __name__ = 'test.many2one_orderby'
    many2one = fields.Many2One('test.many2one_target', 'many2one')


class Many2OneSearch(ModelSQL):
    "Many2One Search"
    __name__ = 'test.many2one_search'
    many2one = fields.Many2One('test.many2one_target', 'many2one')


class Many2OneTree(ModelSQL):
    'Many2One Tree'
    __name__ = 'test.many2one_tree'
    many2one = fields.Many2One('test.many2one_tree', 'many2one')


class Many2OneMPTT(ModelSQL):
    'Many2One MPTT'
    __name__ = 'test.many2one_mptt'
    many2one = fields.Many2One('test.many2one_mptt', 'many2one',
        left='left', right='right')
    left = fields.Integer('Left', required=True)
    right = fields.Integer('Right', required=True)

    @classmethod
    def default_left(cls):
        return 0

    @classmethod
    def default_right(cls):
        return 0


class Many2OneNoForeignKey(ModelSQL):
    "Many2One No Foreign Key"
    __name__ = 'test.many2one_no_foreign_key'
    many2one = fields.Many2One('test.many2one_target_storage', 'many2one')


class Many2OneTargetStorage(ModelStorage):
    "Many2One Target Storage"
    __name__ = 'test.many2one_target_storage'



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import unittest

from trytond.tests.test_tryton import activate_module, with_transaction
from trytond.pool import Pool


class UnionMixinTestCase(unittest.TestCase):
    'Test UnionMixin'

    @classmethod
    def setUpClass(cls):
        activate_module('tests')

    @with_transaction()
    def test_union(self):
        'Test union'
        pool = Pool()
        Union = pool.get('test.union')

        for i in range(1, 4):
            Model = pool.get('test.model.union%s' % i)
            for j in range(3):
                model = Model(name='%s - %s' % (i, j))
                if hasattr(Model, 'optional'):
                    model.optional = 'optional'
                model.save()

        self.assertEqual(len(Union.search([])), 9)
        record, = Union.search([
                ('name', '=', '2 - 2'),
                ])
        self.assertEqual(record.name, '2 - 2')
        self.assertEqual(record.optional, None)

        record, = Union.search([
                ('optional', '=', 'optional'),
                ], limit=1)
        self.assertEqual(record.optional, 'optional')

    @with_transaction()
    def test_union_union(self):
        'Test union of union'
        pool = Pool()
        Union = pool.get('test.union.union')

        for i in range(1, 5):
            Model = pool.get('test.model.union%s' % i)
            for j in range(3):
                model = Model(name='%s - %s' % (i, j))
                model.save()

        self.assertEqual(len(Union.search([])), 12)
        record, = Union.search([
                ('name', '=', '2 - 2'),
                ])
        self.assertEqual(record.name, '2 - 2')
        record, = Union.search([
                ('name', '=', '4 - 1'),
                ])
        self.assertEqual(record.name, '4 - 1')

    @with_transaction()
    def test_union_tree(self):
        'Test union tree'
        pool = Pool()
        Union = pool.get('test.union.tree')
        Model1 = pool.get('test.model.union.tree1')
        Model2 = pool.get('test.model.union.tree2')

        roots = Model1.create([{
                    'name': 'Root1',
                    }, {
                    'name': 'Root2',
                    }, {
                    'name': 'Root3',
                    }])

        Model2.create([{
                    'name': 'Not child',  # To shift ids
                    }, {
                    'name': 'Child1',
                    'parent': roots[1].id,
                    }, {
                    'name': 'Child2',
                    'parent': roots[1].id,
                    }, {
                    'name': 'Child3',
                    'parent': roots[2].id,
                    }])

        uroots = Union.search([('parent', '=', None)],
            order=[('name', 'ASC')])

        self.assertEqual(len(uroots), 4)
        names = [r.name for r in uroots]
        self.assertEqual(names, ['Not child', 'Root1', 'Root2', 'Root3'])
        childs = [len(r.childs) for r in uroots]
        self.assertEqual(childs, [0, 0, 2, 1])
        child_names = sorted((r.name for r in uroots[2].childs))
        self.assertEqual(child_names, ['Child1', 'Child2'])
        self.assertEqual(uroots[3].childs[0].name, 'Child3')


def suite():
    return unittest.TestLoader().loadTestsFromTestCase(UnionMixinTestCase)



# This file is part of Tryton.  The COPYRIGHT file at the toplevel of this
# repository contains the full copyright notices and license terms.
import json

from .field import Field
from ...protocols.jsonrpc import JSONDecoder, JSONEncoder
from ...pool import Pool
from ...tools import grouped_slice


class Dict(Field):
    'Define dict field.'
    _type = 'dict'
    _sql_type = 'TEXT'

    def __init__(self, schema_model, string='', help='', required=False,
            readonly=False, domain=None, states=None, select=False,
            on_change=None, on_change_with=None, depends=None,
            context=None, loading='lazy'):
        super(Dict, self).__init__(string, help, required, readonly, domain,
            states, select, on_change, on_change_with, depends, context,
            loading)
        self.schema_model = schema_model

    def get(self, ids, model, name, values=None):
        dicts = dict((id, None) for id in ids)
        for value in values or []:
            data = value[name]
            if data:
                # If stored as JSON conversion is done on backend
                if isinstance(data, basestring):
                    data = json.loads(data, object_hook=JSONDecoder())
                dicts[value['id']] = data
        return dicts

    def sql_format(self, value):
        if value is None:
            return None
        assert isinstance(value, dict)
        return json.dumps(value, cls=JSONEncoder, separators=(',', ':'))

    def translated(self, name=None, type_='values'):
        "Return a descriptor for the translated value of the field"
        if name is None:
            name = self.name
        if name is None:
            raise ValueError('Missing name argument')
        return TranslatedDict(name, type_)


class TranslatedDict(object):
    'A descriptor for translated values of Dict field'

    def __init__(self, name, type_):
        assert type_ in ['keys', 'values']
        self.name = name
        self.type_ = type_

    def __get__(self, inst, cls):
        if inst is None:
            return self
        pool = Pool()
        schema_model = getattr(cls, self.name).schema_model
        SchemaModel = pool.get(schema_model)

        value = getattr(inst, self.name)
        if not value:
            return value

        domain = []
        if self.type_ == 'values':
            domain = [('type_', '=', 'selection')]

        records = []
        for key_names in grouped_slice(value.keys()):
            records += SchemaModel.search([
                    ('name', 'in', key_names),
                    ] + domain)
        keys = SchemaModel.get_keys(records)

        if self.type_ == 'keys':
            return {k['name']: k['string'] for k in keys}

        elif self.type_ == 'values':
            trans = {k['name']: dict(k['selection']) for k in keys}
            return {k: v if k not in trans else trans[k].get(v, v)
                for k, v in value.iteritems()}



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.transaction import Transaction
from trytond.backend.table import TableHandlerInterface
import logging
import re
import warnings

__all__ = ['TableHandler']

logger = logging.getLogger(__name__)
VARCHAR_SIZE_RE = re.compile('VARCHAR\(([0-9]+)\)')


class TableHandler(TableHandlerInterface):
    def __init__(self, model, module_name=None, history=False):
        super(TableHandler, self).__init__(model,
                module_name=module_name, history=history)
        self._columns = {}
        self._constraints = []
        self._fk_deltypes = {}
        self._indexes = []
        self._model = model

        cursor = Transaction().connection.cursor()
        # Create new table if necessary
        if not self.table_exist(self.table_name):
            if not self.history:
                cursor.execute('CREATE TABLE "%s" '
                    '(id INTEGER PRIMARY KEY AUTOINCREMENT)'
                    % self.table_name)
            else:
                cursor.execute('CREATE TABLE "%s" '
                    '(__id INTEGER PRIMARY KEY AUTOINCREMENT, '
                    'id INTEGER)' % self.table_name)

        self._update_definitions()

    @staticmethod
    def table_exist(table_name):
        cursor = Transaction().connection.cursor()
        cursor.execute("SELECT sql FROM sqlite_master "
            "WHERE type = 'table' AND name = ?",
            (table_name,))
        res = cursor.fetchone()
        if not res:
            return False
        sql, = res

        # Migration from 1.6 add autoincrement

        if 'AUTOINCREMENT' not in sql.upper():
            temp_sql = sql.replace(table_name, '_temp_%s' % table_name)
            cursor.execute(temp_sql)
            cursor.execute('PRAGMA table_info("' + table_name + '")')
            columns = ['"%s"' % column for _, column, _, _, _, _
                    in cursor.fetchall()]
            cursor.execute(('INSERT INTO "_temp_%s" '
                    '(' + ','.join(columns) + ') '
                    'SELECT ' + ','.join(columns) +
                    ' FROM "%s"') % (table_name, table_name))
            cursor.execute('DROP TABLE "%s"' % table_name)
            new_sql = sql.replace('PRIMARY KEY',
                    'PRIMARY KEY AUTOINCREMENT')
            cursor.execute(new_sql)
            cursor.execute(('INSERT INTO "%s" '
                    '(' + ','.join(columns) + ') '
                    'SELECT ' + ','.join(columns) +
                    ' FROM "_temp_%s"') % (table_name, table_name))
            cursor.execute('DROP TABLE "_temp_%s"' % table_name)
        return True

    @staticmethod
    def table_rename(old_name, new_name):
        cursor = Transaction().connection.cursor()
        if (TableHandler.table_exist(old_name)
                and not TableHandler.table_exist(new_name)):
            cursor.execute('ALTER TABLE "%s" RENAME TO "%s"'
                % (old_name, new_name))
        # Rename history table
        old_history = old_name + "__history"
        new_history = new_name + "__history"
        if (TableHandler.table_exist(old_history)
                and not TableHandler.table_exist(new_history)):
            cursor.execute('ALTER TABLE "%s" RENAME TO "%s"'
                % (old_history, new_history))

    def column_exist(self, column_name):
        return column_name in self._columns

    def _recreate_table(self, new_columns):
        transaction = Transaction()
        database = transaction.database
        cursor = transaction.connection.cursor()
        temp_table = '_temp_%s' % self.table_name
        TableHandler.table_rename(self.table_name, temp_table)
        new_table = TableHandler(self._model, history=self.history)
        columns, old_columns = [], []
        for column, values in self._columns.iteritems():
            typname = new_columns.get(column, {}).get(
                'typname', values['typname'])
            size = new_columns.get(column, {}).get('size', values['size'])
            new_column = new_columns.get(column, {}).get('name', column)
            new_table._add_raw_column(
                new_column, database.sql_type(typname), field_size=size)
            columns.append(new_column)
            old_columns.append(column)
        cursor.execute(('INSERT INTO "%s" (' +
                ','.join('"%s"' % x for x in columns) +
                ') SELECT ' +
                ','.join('"%s"' % x for x in old_columns) + ' ' +
                'FROM "%s"') % (self.table_name, temp_table))
        cursor.execute('DROP TABLE "%s"' % temp_table)
        self._update_definitions()

    def column_rename(self, old_name, new_name, exception=False):
        if self.column_exist(old_name) and \
                not self.column_exist(new_name):
            self._recreate_table({old_name: {'name': new_name}})
        elif exception and self.column_exist(new_name):
            raise Exception('Unable to rename column %s.%s to %s.%s: '
                '%s.%s already exist!'
                % (self.table_name, old_name, self.table_name, new_name,
                    self.table_name, new_name))

    def _update_definitions(self, columns=None, indexes=None):
        if columns is None and indexes is None:
            columns = indexes = True
        cursor = Transaction().connection.cursor()
        # Fetch columns definitions from the table
        if columns:
            cursor.execute('PRAGMA table_info("' + self.table_name + '")')
            self._columns = {}
            for _, column, type_, notnull, hasdef, _ in cursor.fetchall():
                column = re.sub(r'^\"|\"$', '', column)
                match = re.match(r'(\w+)(\((.*?)\))?', type_)
                if match:
                    typname = match.group(1).upper()
                    size = match.group(3) and int(match.group(3)) or 0
                else:
                    typname = type_.upper()
                    size = -1
                self._columns[column] = {
                    'notnull': notnull,
                    'hasdef': hasdef,
                    'size': size,
                    'typname': typname,
                }

        # Fetch indexes defined for the table
        if indexes:
            try:
                cursor.execute('PRAGMA index_list("' + self.table_name + '")')
            except IndexError:  # There is sometimes IndexError
                cursor.execute('PRAGMA index_list("' + self.table_name + '")')
            self._indexes = [l[1] for l in cursor.fetchall()]

    @property
    def _field2module(self):
        cursor = Transaction().connection.cursor()
        cursor.execute('SELECT f.name, f.module '
            'FROM ir_model_field f '
            'JOIN ir_model m on (f.model=m.id) '
            'WHERE m.model = ?',
            (self.object_name,))
        return dict(cursor)

    def alter_size(self, column_name, column_type):
        self._recreate_table({column_name: {'size': column_type}})

    def alter_type(self, column_name, column_type):
        self._recreate_table({column_name: {'typname': column_type}})

    def db_default(self, column_name, value):
        warnings.warn('Unable to set default on column with SQLite backend')

    def add_column(self, column_name, sql_type, default=None, comment=''):
        database = Transaction().database
        column_type = database.sql_type(sql_type)
        match = VARCHAR_SIZE_RE.match(sql_type)
        field_size = int(match.group(1)) if match else None

        self._add_raw_column(column_name, column_type, default, field_size,
            comment)

    def _add_raw_column(self, column_name, column_type, default=None,
            field_size=None, string=''):
        if self.column_exist(column_name):
            base_type = column_type[0].upper()
            if base_type != self._columns[column_name]['typname']:
                if (self._columns[column_name]['typname'], base_type) in [
                        ('VARCHAR', 'TEXT'),
                        ('TEXT', 'VARCHAR'),
                        ('DATE', 'TIMESTAMP'),
                        ('INTEGER', 'FLOAT'),
                        ]:
                    self.alter_type(column_name, base_type)
                else:
                    logger.warning(
                        'Unable to migrate column %s on table %s '
                        'from %s to %s.',
                        column_name, self.table_name,
                        self._columns[column_name]['typname'], base_type)

            if (base_type == 'VARCHAR'
                    and self._columns[column_name]['typname'] == 'VARCHAR'):
                # Migrate size
                from_size = self._columns[column_name]['size']
                if field_size is None:
                    if from_size > 0:
                        self.alter_size(column_name, base_type)
                elif from_size == field_size:
                    pass
                elif from_size and from_size < field_size:
                    self.alter_size(column_name, column_type[1])
                else:
                    logger.warning(
                        'Unable to migrate column %s on table %s '
                        'from varchar(%s) to varchar(%s).',
                        column_name, self.table_name,
                        from_size if from_size and from_size > 0 else "",
                        field_size)
            return

        cursor = Transaction().connection.cursor()
        column_type = column_type[1]
        cursor.execute(('ALTER TABLE "%s" ADD COLUMN "%s" %s') %
                       (self.table_name, column_name, column_type))

        if default:
            # check if table is non-empty:
            cursor.execute('SELECT 1 FROM "%s" limit 1' % self.table_name)
            if cursor.fetchone():
                # Populate column with default values:
                cursor.execute('UPDATE "' + self.table_name + '" '
                    'SET "' + column_name + '" = ?', (default(),))

        self._update_definitions(columns=True)

    def add_fk(self, column_name, reference, on_delete=None):
        warnings.warn('Unable to add foreign key with SQLite backend')

    def drop_fk(self, column_name, table=None):
        warnings.warn('Unable to drop foreign key with SQLite backend')

    def index_action(self, column_name, action='add', table=None):
        if isinstance(column_name, basestring):
            column_name = [column_name]
        index_name = self.convert_name(
            self.table_name + "_" + '_'.join(column_name) + "_index")

        cursor = Transaction().connection.cursor()
        if action == 'add':
            if index_name in self._indexes:
                return
            cursor.execute('CREATE INDEX "' + index_name + '" '
                'ON "' + self.table_name + '" ( ' +
                ','.join('"' + x + '"' for x in column_name) +
                ')')
            self._update_definitions(indexes=True)
        elif action == 'remove':
            if len(column_name) == 1:
                if self._field2module.get(column_name[0],
                        self.module_name) != self.module_name:
                    return

            if index_name in self._indexes:
                cursor.execute('DROP INDEX "%s" ' % (index_name,))
                self._update_definitions(indexes=True)
        else:
            raise Exception('Index action not supported!')

    def not_null_action(self, column_name, action='add'):
        if not self.column_exist(column_name):
            return

        if action == 'add':
            warnings.warn('Unable to set not null with SQLite backend')
        elif action == 'remove':
            warnings.warn('Unable to remove not null with SQLite backend')
        else:
            raise Exception('Not null action not supported!')

    def add_constraint(self, ident, constraint, exception=False):
        warnings.warn('Unable to add constraint with SQLite backend')

    def drop_constraint(self, ident, exception=False, table=None):
        warnings.warn('Unable to drop constraint with SQLite backend')

    def drop_column(self, column_name, exception=False):
        if not self.column_exist(column_name):
            return
        transaction = Transaction()
        database = transaction.database
        cursor = transaction.connection.cursor()
        temp_table = '_temp_%s' % self.table_name
        try:
            TableHandler.table_rename(self.table_name, temp_table)
        except Exception:
            if exception:
                raise
            logger.warning('Unable to drop column: Error renaming '
                'table %s to temp!', self.table_name)
        new_table = TableHandler(self._model, history=self.history)
        for name, values in self._columns.iteritems():
            if name != column_name:
                typname = values['typname']
                size = values['size']
                new_table._add_raw_column(
                    name, database.sql_type(typname), field_size=size)
        columns_name = [x for x in new_table._columns.keys()]
        cursor.execute(('INSERT INTO "%s" (' +
                        ','.join('"%s"' % c for c in columns_name) +
                        ') SELECT ' +
                        ','.join('"%s"' % c for c in columns_name) + ' ' +
                        'FROM "%s"') % (self.table_name, temp_table))
        try:
            cursor.execute('DROP TABLE "%s"' % temp_table)
        except Exception:
            if exception:
                raise
            logger.warning('Unable to drop temporary table %s',
                self.temp_table)
        self._update_definitions()

    @staticmethod
    def drop_table(model, table, cascade=False):
        cursor = Transaction().connection.cursor()
        cursor.execute('DELETE from ir_model_data where '
            'model = \'%s\'' % model)

        query = 'DROP TABLE "%s"' % table
        if cascade:
            query = query + ' CASCADE'
        cursor.execute(query)



# -*- coding: utf-8 -*-
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import unittest
import datetime
from trytond.tests.test_tryton import activate_module, with_transaction
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.exceptions import UserError


class SequenceTestCase(unittest.TestCase):
    'Test Sequence'

    @classmethod
    def setUpClass(cls):
        activate_module('tests')

    @staticmethod
    def get_model():
        pool = Pool()
        return pool.get('ir.sequence')

    @with_transaction()
    def test_incremental(self):
        'Test incremental'
        Sequence = self.get_model()

        sequence, = Sequence.create([{
                    'name': 'Test incremental',
                    'code': 'test',
                    'prefix': '',
                    'suffix': '',
                    'type': 'incremental',
                    }])
        self.assertEqual(sequence.number_next, 1)
        self.assertEqual(Sequence.get_id(sequence), '1')

        Sequence.write([sequence], {
                'number_increment': 10,
                })
        self.assertEqual(sequence.number_next, 2)
        self.assertEqual(Sequence.get_id(sequence), '2')
        self.assertEqual(Sequence.get_id(sequence), '12')

        Sequence.write([sequence], {
                'padding': 3,
                })
        self.assertEqual(sequence.number_next, 22)
        self.assertEqual(Sequence.get_id(sequence), '022')

    @with_transaction()
    def test_decimal_timestamp(self):
        'Test Decimal Timestamp'
        Sequence = self.get_model()

        sequence, = Sequence.create([{
                    'name': 'Test decimal timestamp',
                    'code': 'test',
                    'prefix': '',
                    'suffix': '',
                    'type': 'decimal timestamp',
                    }])
        timestamp = Sequence.get_id(sequence)
        self.assertEqual(timestamp, str(sequence.last_timestamp))

        self.assertEqual(sequence.number_next, None)

        self.assertNotEqual(Sequence.get_id(sequence), timestamp)

        next_timestamp = Sequence._timestamp(sequence)
        self.assertRaises(UserError, Sequence.write, [sequence], {
                'last_timestamp': next_timestamp + 100,
                })

    @with_transaction()
    def test_hexadecimal_timestamp(self):
        'Test Hexadecimal Timestamp'
        Sequence = self.get_model()

        sequence, = Sequence.create([{
                    'name': 'Test hexadecimal timestamp',
                    'code': 'test',
                    'prefix': '',
                    'suffix': '',
                    'type': 'hexadecimal timestamp',
                    }])
        timestamp = Sequence.get_id(sequence)
        self.assertEqual(timestamp,
            hex(int(sequence.last_timestamp))[2:].upper())

        self.assertEqual(sequence.number_next, None)

        self.assertNotEqual(Sequence.get_id(sequence), timestamp)

        next_timestamp = Sequence._timestamp(sequence)
        self.assertRaises(UserError, Sequence.write, [sequence], {
                'last_timestamp': next_timestamp + 100,
                })

    @with_transaction()
    def test_prefix_suffix(self):
        'Test prefix/suffix'
        Sequence = self.get_model()

        sequence, = Sequence.create([{
                    'name': 'Test incremental',
                    'code': 'test',
                    'prefix': 'prefix/',
                    'suffix': '/suffix',
                    'type': 'incremental',
                    }])
        self.assertEqual(Sequence.get_id(sequence),
            'prefix/1/suffix')

        Sequence.write([sequence], {
                'prefix': '${year}-${month}-${day}/',
                'suffix': '/${day}.${month}.${year}',
                })
        with Transaction().set_context(date=datetime.date(2010, 8, 15)):
            self.assertEqual(Sequence.get_id(sequence),
                '2010-08-15/2/15.08.2010')


class SequenceStrictTestCase(SequenceTestCase):
    "Test Sequence Strict"

    @staticmethod
    def get_model():
        pool = Pool()
        return pool.get('ir.sequence.strict')


def suite():
    suite_ = unittest.TestSuite()
    suite_.addTests(unittest.TestLoader().loadTestsFromTestCase(
            SequenceTestCase))
    suite_.addTests(unittest.TestLoader().loadTestsFromTestCase(
            SequenceStrictTestCase))
    return suite_
